module hol.

kind    tm          type.
kind    ty          type.
kind    nprf        type.

type    abs         (tm -> tm) -> tm.
type    app         tm -> tm -> tm.

type    and         tm -> tm -> tm.
type    or          tm -> tm -> tm.
type    imp         tm -> tm -> tm.
type    neg         tm -> tm.
type    forall      ty -> (tm -> tm) -> tm.
type    exists      ty -> (tm -> tm) -> tm.
type    false       tm.

infixl  or          4.
infixl  and         5.
infixr  imp         6.

type    and_i       nprf -> nprf -> nprf.
type    and_e1      nprf -> nprf.
type    and_e2      nprf -> nprf.
type    or_i1       nprf -> nprf.
type    or_i2       nprf -> nprf.
type    or_e        nprf -> (nprf -> nprf) -> (nprf -> nprf) -> nprf.
type    imp_i       (nprf -> nprf) -> nprf.
type    imp_e       nprf -> nprf -> nprf.
type    neg_i       (nprf -> nprf) -> nprf.
type    neg_e       nprf -> nprf -> nprf.
type    exists_i    ty -> tm -> nprf -> nprf.
type    exists_e    nprf -> (tm -> nprf -> nprf) -> nprf.
type    forall_i    ty -> (tm -> nprf) -> nprf.
type    forall_e    ty -> tm -> nprf -> nprf.
type    false_i     nprf -> nprf.
type    convert     tm -> tm -> nprf -> nprf.
type    redex       tm -> tm -> o.
type    red1        tm -> tm -> o.
type    conv        tm -> tm -> o.

type    form        ty.
type    -->         ty -> ty -> ty.
type    #t          tm -> ty -> o.
type    #p          nprf -> tm -> o.
type    #           nprf -> tm -> o.

infix   -->         4.
infix   #t          3.
infix   #p          3.
infix   #           3.

redex (app (abs M) N)  (M N).
redex (abs X\(app M X)) M.

red1 M N :- redex M N.
red1 (abs M) (abs N) :- pi X\ (red1 (M X) (N X)).
red1 (app M N) (app P N) :- red1 M P.
red1 (app M N) (app M P) :- red1 N P.
red1 (A and B) (C and B) :- red1 A C.
red1 (A and B) (A and C) :- red1 B C.
red1 (A or B) (C or B) :- red1 A C.
red1 (A or B) (A or C) :- red1 B C.
red1 (A imp B) (C imp B) :- red1 A C.
red1 (A imp B) (A imp C) :- red1 B C.
red1 (neg A) (neg B) :- red1 A B.
red1 (forall S A) (forall S B) :- pi X\ (red1 (A X) (B X)).
red1 (exists S A) (exists S B) :- pi X\ (red1 (A X) (B X)).

conv M N :- red1 M P.
conv M M.
conv M N :- conv N M.
conv M N :- conv M P, conv P N.

(abs M) #t (R --> S) :- pi X\ ((X #t R) => ((M X) #t S)).
(app M N) #t S :- M #t (R --> S), N #t R.
(A and B) #t form :- A #t form, B #t form.
(A or B) #t form :- A #t form, B #t form.
(A imp B) #t form :- A #t form, B #t form.
(neg A) #t form :- A #t form.
(exists S A) #t form :- pi X\ ((X #t S) => ((A X) #t form)).
(forall S A) #t form :- pi X\ ((X #t S) => ((A X) #t form)).

(and_i P1 P2) #p (A and B) :- P1 #p A, P2 #p B.
(or_i1 P) #p (A or B) :- P #p A.
(or_i2 P) #p (A or B) :- P #p B.
(imp_i P) #p (A imp B) :- pi PA\ ((PA #p A) => ((P PA) #p B)).
(neg_i P) #p (neg A) :- pi PA\ ((PA #p A) => ((P PA) #p false)).
(false_i P) #p A :- P #p false.

(and_e1 P) #p A :- P #p (A and B), B #t form.
(and_e2 P) #p B :- P #p (A and B), A #t form.

(or_e P P1 P2) #p C :- P #p (A or B), (A or B) #t form,
                       pi PA\ ((PA #p A) => ((P1 PA) #p C)),
                       pi PB\ ((PB #p B) => ((P2 PB) #p C)).

(imp_e P1 P2) #p B :- P1 #p A, A #t form,
                      P2 #p (A imp B).

(neg_e P1 P2) #p false :- P1 # A, A #t form,
                          P2 #p (neg A).

(exists_i S T P) #p (exists S A) :- T #t S, P #p (A T).

(forall_i S P) #p (forall S A) :- pi Y\ ((Y #t S) => ((P Y) #p (A Y))).

(forall_e S T P) #p (A T) :- T #t S, (forall S A) #t form,
                              P #p (forall S A).
(exists_e P1 P2) #p B :-
   P1 #p (exists S A), (exists S A) #t form,
   pi Y\ (pi P\ ((Y #t S) => ((P #p (A Y)) => ((P2 Y P) #p B)))).

(convert A B P) #p A :- B #t form, P #p B, conv A B.

P # A :- A #t form, P #p A.
